/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmDlgWin.h"

#include "Emulator.h"			// gInstance
#include "Miscellaneous.h"		// ReplaceString
#include "Platform.h"			// Platform::GetString

#include <tchar.h>				// _tcsinc, etc.

#include "resource.h"			// IDD_ROM_TRANSFER_QUERY, IDD_ROM_TRANSFER_PROGRESS

typedef char* POSITION;
typedef vector<char>	FilterList;

static UINT				PrvFromDlgID (EmDlgID id);
static EmDlgID			PrvToDlgID (UINT id);
static UINT				PrvFromDlgItemID (EmDlgItemID id);
static EmDlgItemID		PrvToDlgItemID (UINT id);

static FilterList		PrvConvertTypeList (const EmFileTypeList& typeList);
static void				PrvAddFilter (FilterList&, const char*, const char*);

static LRESULT CALLBACK	PrvOpenMultiFileHook(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static UINT CALLBACK	PrvOpenMultiFileProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static string			PrvGetNextPathName (OPENFILENAME& ofn, POSITION& pos);
static POSITION			PrvGetStartPosition (OPENFILENAME& ofn);

static Bool				PrvIsPushButton (HWND hWnd);
static Bool				PrvIsCheckBox (HWND hWnd);
static Bool				PrvIsRadioButton (HWND hWnd);
static Bool				PrvIsProgressBar (HWND hWnd);
static Bool				PrvIsButton (HWND hWnd);
static Bool				PrvIsComboBox (HWND hWnd);
static Bool				PrvIsListBox (HWND hWnd);
static Bool				PrvIsEdit (HWND hWnd);
static Bool				PrvIsClass (HWND hWnd, const char* className);

static WNDPROC			g_lpfnDialogProc;

// Array for mapping between MSDEV Resource Editor-assigned resource IDs
// and our own internal, cross-platform IDs.

#define ID_UNMAPPED	((UINT) -1)

struct
{
	EmDlgID		fMyID;
	UINT		fWinID;
} kDlgIDMap[] =
{
	{ kDlgAboutBox,					ID_UNMAPPED },
	{ kDlgSessionNew,				ID_UNMAPPED },
	{ kDlgSessionSave,				ID_UNMAPPED },
	{ kDlgHordeNew,					IDD_NEW_HORDE },
	{ kDlgDatabaseImportProgress,	ID_UNMAPPED },
	{ kDlgDatabaseExport,			ID_UNMAPPED },
	{ kDlgROMTransferQuery,			IDD_ROM_TRANSFER_QUERY },
	{ kDlgROMTransferProgress,		IDD_ROM_TRANSFER_PROGRESS },
	{ kDlgGremlinControl,			ID_UNMAPPED },
	{ kDlgEditPreferences,			ID_UNMAPPED },
	{ kDlgEditLogging,				IDD_EDIT_LOGGING_OPTIONS },
	{ kDlgEditDebugging,			ID_UNMAPPED },
	{ kDlgEditSkins,				IDD_EDIT_SKINS },
	{ kDlgCommonDialog,				ID_UNMAPPED },
	{ kDlgSaveBound,				IDD_SAVE_BOUND }
};

struct
{
	EmDlgItemID	fMyID;
	UINT		fWinID;
} kDlgItemIDMap[] =
{
	{ kDlgItemNone,					0 },

	{ kDlgItemOK,					IDOK },
	{ kDlgItemCancel,				IDCANCEL },

	{ kDlgItemAbtAppName,			ID_UNMAPPED },
	{ kDlgItemAbtURLPalmWeb,		ID_UNMAPPED },
	{ kDlgItemAbtURLPalmMail,		ID_UNMAPPED },
	{ kDlgItemAbtURLWindowsWeb,		ID_UNMAPPED },
	{ kDlgItemAbtURLWindowsMail,	ID_UNMAPPED },
	{ kDlgItemAbtURLMacWeb,			ID_UNMAPPED },
	{ kDlgItemAbtURLMacMail,		ID_UNMAPPED },
	{ kDlgItemAbtURLUAEWeb,			ID_UNMAPPED },
	{ kDlgItemAbtURLUAEMail,		ID_UNMAPPED },
	{ kDlgItemAbtURLQNXWeb,			ID_UNMAPPED },
	{ kDlgItemAbtURLQNXMail,		ID_UNMAPPED },

	{ kDlgItemNewDevice,			ID_UNMAPPED },
	{ kDlgItemNewSkin,				ID_UNMAPPED },
	{ kDlgItemNewMemorySize,		ID_UNMAPPED },
	{ kDlgItemNewROM,				ID_UNMAPPED },
	{ kDlgItemNewROMBrowse,			ID_UNMAPPED },

	{ kDlgItemHrdAppList,			IDC_HORDE_APP_LIST },
	{ kDlgItemHrdStart,				IDC_HORDE_START_NUMBER },
	{ kDlgItemHrdStop,				IDC_HORDE_STOP_NUMBER },
	{ kDlgItemHrdCheckSwitch,		IDC_HORDE_SWITCH_CHECK },
	{ kDlgItemHrdCheckSave,			IDC_HORDE_SAVE_CHECK },
	{ kDlgItemHrdCheckStop,			IDC_HORDE_STOP_CHECK },
	{ kDlgItemHrdDepthSwitch,		IDC_HORDE_SWITCH_EVENTS },
	{ kDlgItemHrdDepthSave,			IDC_HORDE_SAVE_EVENTS },
	{ kDlgItemHrdDepthStop,			IDC_HORDE_STOP_EVENTS },
	{ kDlgItemHrdLogging,			IDC_HORDE_LOGGING },

	{ kDlgItemExpDbList,			ID_UNMAPPED },

	{ kDlgItemDlqInstructions,		IDC_INSTRUCTIONS },
	{ kDlgItemDlqPortList,			IDC_PORT_LIST },
	{ kDlgItemDlqBaudList,			IDC_SPEED_LIST },

	{ kDlgItemDlpMessage,			IDC_TRANSFER_MESSAGE },
	{ kDlgItemDlpProgress,			IDC_PROGRESS },

	{ kDlgItemLogNormal,			IDC_LOG_NORMAL },
	{ kDlgItemLogGremlins,			IDC_LOG_GREMLINS },
#define DEFINE_BUTTON_ID_MAPPING(name)	\
	{ kDlgItemLog##name,			IDC_##name },
	FOR_EACH_LOG_PREF(DEFINE_BUTTON_ID_MAPPING)

	{ kDlgItemSknSkinList,			IDC_SKINLIST },
	{ kDlgItemSknDoubleScale,		IDC_SCALE },
	{ kDlgItemSknWhiteBackground,	IDC_WHITEBACKGROUND },

	{ kDlgItemBndSaveROM,			IDC_BIND_ROM },
	{ kDlgItemBmdSaveRAM,			IDC_BIND_RAM },

	{ kDlgItemLast,					ID_UNMAPPED }
};


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunGetFile
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to load
 *				the file.
 *
 * PARAMETERS:	typeList - types of files user is allowed to select.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunGetFile (	EmFileRef& result,
									const string& prompt,
									const EmDirRef& defaultPath,
									const EmFileTypeList& filterList)
{
	FilterList		filters = ::PrvConvertTypeList (filterList);

	char			fileName[MAX_PATH] = {0};

	string			initialDir;

	if (defaultPath.IsSpecified ())
		initialDir = defaultPath.GetFullPath ();

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hwndOwner 		= NULL; 	// hWnd; -- Don't do this, or dialog will be too small!
	reply.hInstance 		= gInstance;
	reply.lpstrFilter		= &filters[0];
	reply.nFilterIndex		= 1;
	reply.lpstrFile 		= fileName;
	reply.nMaxFile			= MAX_PATH;
	reply.lpstrInitialDir	= initialDir.c_str ();
	reply.lpstrTitle		= prompt.empty() ? NULL : prompt.c_str ();
	reply.Flags 			= OFN_ENABLEHOOK | OFN_EXPLORER | OFN_FILEMUSTEXIST;
	reply.lpfnHook			= &::CenterDlgProc;

	BOOL	ok = ::GetOpenFileName (&reply);

	if (ok)
	{
		result = EmFileRef (fileName);
	}

	return ok ? kDlgItemOK : kDlgItemCancel;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunGetFileList
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to load
 *				the file.
 *
 * PARAMETERS:	typeList - types of files user is allowed to select.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunGetFileList (	EmFileRefList& results,
										const string& prompt,
										const EmDirRef& defaultPath,
										const EmFileTypeList& filterList)
{
	FilterList		filters = ::PrvConvertTypeList (filterList);

	string			initialDir;

	if (defaultPath.IsSpecified ())
		initialDir = defaultPath.GetFullPath ();

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hwndOwner 		= NULL; 	// hWnd; -- Don't do this, or dialog will be too small!
	reply.hInstance 		= gInstance;
	reply.lpstrFilter		= &filters[0];
	reply.nFilterIndex		= 1;
	reply.lpstrFile 		= (LPTSTR) malloc (MAX_PATH);
	reply.nMaxFile			= MAX_PATH;
	reply.lpstrInitialDir	= initialDir.c_str ();
	reply.lpstrTitle		= prompt.empty() ? NULL : prompt.c_str ();
	reply.Flags 			= OFN_ENABLEHOOK | OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_ALLOWMULTISELECT;
	reply.lpfnHook			= &::PrvOpenMultiFileProc;

	reply.lpstrFile[0] = 0;

	BOOL	ok = ::GetOpenFileName (&reply);

	if (ok)
	{
		POSITION	pos = ::PrvGetStartPosition (reply);
		while (pos)
		{
			string	pathName = ::PrvGetNextPathName (reply, pos);
			results.push_back (EmFileRef (pathName));
		}
	}

	free (reply.lpstrFile);

	return ok ? kDlgItemOK : kDlgItemCancel;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunPutFile
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to save
 *				the file.
 *
 * PARAMETERS:	defName - default name for the file to be saved.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunPutFile (	EmFileRef& result,
									const string& prompt,
									const EmDirRef& defaultPath,
									const EmFileTypeList& filterList,
									const string& defName)
{
	FilterList		filters = ::PrvConvertTypeList (filterList);

	char			fileName[MAX_PATH] = {0};

	if (!defName.empty ())
		strcpy (fileName, defName.c_str ());

	string			initialDir;

	if (defaultPath.IsSpecified ())
		initialDir = defaultPath.GetFullPath ();

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hwndOwner 		= NULL; 	// hWnd; -- Don't do this, or dialog will be too small!
	reply.hInstance 		= gInstance;
	reply.lpstrFilter		= &filters[0];
	reply.nFilterIndex		= 1;
	reply.lpstrFile 		= fileName;
	reply.nMaxFile			= MAX_PATH;
	reply.lpstrInitialDir	= initialDir.c_str ();
	reply.lpstrTitle		= prompt.empty() ? NULL : prompt.c_str ();
	reply.Flags 			= OFN_ENABLEHOOK | OFN_OVERWRITEPROMPT | OFN_EXPLORER;
	reply.lpstrDefExt		= "psf";
	reply.lpfnHook			= &::CenterDlgProc;

	BOOL	ok = ::GetSaveFileName (&reply);

	if (ok)
	{
		result = EmFileRef (fileName);
	}

	return ok ? kDlgItemOK : kDlgItemCancel;
}


/***********************************************************************
 *
 * FUNCTION:	HostRunDialog
 *
 * DESCRIPTION:	Common routine that handles the creation of a dialog,
 *				initializes it (via the dialog handler), fetches events,
 *				handles events (via the dialog handler), and closes
 *				the dialog.
 *
 * PARAMETERS:	fn - the custom dialog handler
 *				userData - custom data passed back to the dialog handler.
 *				dlgID - ID of dialog to create.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

static BOOL CALLBACK PrvHostRunDialogProc (	HWND hWnd,
											UINT msg,
											WPARAM wParam,
											LPARAM lParam)
{
	// Except in response to the WM_INITDIALOG message, the dialog box procedure
	// should return nonzero if it processes the message, and zero if it does not.
	// In response to a WM_INITDIALOG message, the dialog box procedure should
	// return zero if it calls the SetFocus function to set the focus to one of
	// the controls in the dialog box. Otherwise, it should return nonzero, in
	// which case the system sets the focus to the first control in the dialog
	// box that can be given the focus.

	// Remember our custom data for later.

	if (msg == WM_INITDIALOG)
	{
		::SetWindowLong (hWnd, GWL_USERDATA, lParam);
	}

	EmDlgContext&	context = *(EmDlgContext*) ::GetWindowLong (hWnd, GWL_USERDATA);

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			context.fDlg = (EmDlgRef) hWnd;

			// Center the window.

			::CenterWindow (hWnd);

			// Call the user's Init function.

			if (context.Init () == kDlgResultClose)
			{
				::EndDialog (hWnd, kDlgItemNone);
				return TRUE;
			}

			// The dialog box procedure should return TRUE to direct the
			// system to set the keyboard focus to the control specified
			// by wParam. Otherwise, it should return FALSE to prevent
			// the system from setting the default keyboard focus.

			return TRUE;
		}

		case WM_COMMAND:
		{
			UINT		wItemID = LOWORD (wParam);
			EmDlgItemID	itemID = ::PrvToDlgItemID (wItemID);

			if (context.Event (itemID) == kDlgResultClose)
			{
				::EndDialog (hWnd, itemID);
			}

			return TRUE;
		}

		case WM_TIMER:
		{
			if (context.Idle () == kDlgResultClose)
			{
				::EndDialog (hWnd, kDlgItemOK);
			}

			return TRUE;
		}

		case WM_CLOSE:
		{
			EmDlg::HostStopIdling (context);
			return FALSE;
		}
	}

	// Typically, the dialog box procedure should return TRUE if it
	// processed the message, and FALSE if it did not. If the dialog
	// box procedure returns FALSE, the dialog manager performs the
	// default dialog operation in response to the message.

	return FALSE;
}

static BOOL CALLBACK PrvHostRunPropertySheetProc (	HWND hDlg,
													UINT msg,
													WPARAM wParam,
													LPARAM lParam)
{
	UNUSED_PARAM(wParam)

	// Except in response to the WM_INITDIALOG message, the dialog box procedure
	// should return nonzero if it processes the message, and zero if it does not.
	// In response to a WM_INITDIALOG message, the dialog box procedure should
	// return zero if it calls the SetFocus function to set the focus to one of
	// the controls in the dialog box. Otherwise, it should return nonzero, in
	// which case the system sets the focus to the first control in the dialog
	// box that can be given the focus.

	// Remember our custom data for later.

	if (msg == WM_INITDIALOG)
	{
		PROPSHEETPAGE*	page = (PROPSHEETPAGE*) lParam;
		::SetWindowLong (hDlg, GWL_USERDATA, page->lParam);
	}

	EmDlgContext&	context = *(EmDlgContext*) ::GetWindowLong (hDlg, GWL_USERDATA);

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			context.fDlg = (EmDlgRef) hDlg;

			// Center the window.

			::CenterWindow (::GetParent(hDlg));

			// Call the user's Init function.

			if (context.Init () == kDlgResultClose)
			{
			//	::EndDialog (hDlg, kDlgItemNone);
				PropSheet_PressButton (hDlg, PSBTN_CANCEL);
				return TRUE;
			}

			// The dialog box procedure should return TRUE to direct the
			// system to set the keyboard focus to the control specified
			// by wParam. Otherwise, it should return FALSE to prevent
			// the system from setting the default keyboard focus.

			return TRUE;
		}

		case WM_COMMAND:
		{
			UINT		wItemID = LOWORD (wParam);
			EmDlgItemID	itemID = ::PrvToDlgItemID (wItemID);

			if (context.Event (itemID) == kDlgResultClose)
			{
			//	::EndDialog (hDlg, itemID);
				PropSheet_PressButton (hDlg, PSBTN_CANCEL);
			}

			return TRUE;
		}

		case WM_TIMER:
		{
			if (context.Idle () == kDlgResultClose)
			{
			//	::EndDialog (hDlg, kDlgItemOK);
				PropSheet_PressButton (hDlg, PSBTN_CANCEL);
			}

			return TRUE;
		}

		case WM_CLOSE:
		{
			EmDlg::HostStopIdling (context);
			return FALSE;
		}

		case WM_NOTIFY:
		{
			switch (((NMHDR*) lParam)->code) 
			{
				case PSN_SETACTIVE:
				{
					// Returns zero to accept the activation, or -1 to activate 
					// the next or the previous page (depending on whether the 
					// user clicked the Next or Back button). To set the activation 
					// to a particular page, return the resource identifier of the page.

					context.PanelEnter ();
					::SetWindowLong (hDlg, DWL_MSGRESULT, 0);
					return TRUE;
				}

				case PSN_KILLACTIVE:
				{
					// Returns TRUE to prevent the page from losing the activation,
					// or FALSE to allow it.

					context.PanelExit ();
					::SetWindowLong (hDlg, DWL_MSGRESULT, FALSE);
					return TRUE;
				}

				case PSN_APPLY:
				{
					// Returns the PSNRET_INVALID_NOCHANGEPAGE value to prevent the
					// changes from taking effect and to return the focus to the page,
					// or the PSNRET_NOERROR value to accept the changes and allow the
					// property sheet to be destroyed.

					context.Event (kDlgItemOK);
					::SetWindowLong (hDlg, DWL_MSGRESULT, PSNRET_NOERROR);
					return TRUE;
				}

				case PSN_RESET:
				{
					context.Event (kDlgItemCancel);

					// No return value.

					return TRUE;
				}
			}
		}
	}

	// Typically, the dialog box procedure should return TRUE if it
	// processed the message, and FALSE if it did not. If the dialog
	// box procedure returns FALSE, the dialog manager performs the
	// default dialog operation in response to the message.

	return FALSE;
}


EmDlgItemID PrvRunEditLoggingOptionsDialog (EmDlgContext& context)
{
#if 0
	PROPSHEETPAGE	psp[2];
	::ZeroMemory (psp, sizeof (psp));

	EmDlgContext	contexts[2] = {context, context};

	contexts[0].fPanelID = kDlgPanelLogNormal;
	contexts[1].fPanelID = kDlgPanelLogGremlins;

	psp[0].dwSize		= sizeof (PROPSHEETPAGE);
	psp[0].dwFlags		= PSP_DEFAULT | PSP_USETITLE;
	psp[0].hInstance	= gInstance;
	psp[0].pszTemplate	= MAKEINTRESOURCE (IDD_EDIT_LOGGING_OPTIONS);
	psp[0].pszTitle 	= MAKEINTRESOURCE (IDS_LOGGINGNORMAL);
	psp[0].pfnDlgProc	= &::PrvHostRunPropertySheetProc;
	psp[0].lParam		= (LPARAM) &contexts[0];

	psp[1].dwSize		= sizeof (PROPSHEETPAGE);
	psp[1].dwFlags		= PSP_DEFAULT | PSP_USETITLE;
	psp[1].hInstance	= gInstance;
	psp[1].pszTemplate	= MAKEINTRESOURCE (IDD_EDIT_LOGGING_OPTIONS);
	psp[1].pszTitle 	= MAKEINTRESOURCE (IDS_LOGGINGGREMLINS);
	psp[1].pfnDlgProc	= &::PrvHostRunPropertySheetProc;
	psp[1].lParam		= (LPARAM) &contexts[1];

	PROPSHEETHEADER psh;
	::ZeroMemory (&psh, sizeof (psh));

	psh.dwSize			= sizeof (psh);
	psh.dwFlags 		= PSH_PROPSHEETPAGE | PSH_NOAPPLYNOW;
	psh.hwndParent		= NULL;
	psh.hInstance		= gInstance;
	psh.pszCaption		= MAKEINTRESOURCE (IDS_LOGGINGOPTIONS);
	psh.nPages			= countof (psp);
	psh.nStartPage		= *((EmDlgPanelID*) context.fUserData) == kDlgPanelLogNormal ? 0 : 1;
	psh.ppsp			= psp;

	(void) ::PropertySheet (&psh);
#endif

	return kDlgItemOK;
}


EmDlgItemID EmDlg::HostRunDialog (EmDlgFn fn, void* userData, EmDlgID dlgID)
{
	EmDlgContext	context;

	context.fFn = fn;
	context.fDlgID = dlgID;
	context.fUserData = userData;

	int	windowsDlgID = ::PrvFromDlgID (dlgID);
	assert (windowsDlgID != 0);

	int choice = ::DialogBoxParam (gInstance, MAKEINTRESOURCE (windowsDlgID),
								NULL, &::PrvHostRunDialogProc, (LPARAM) &context);

	return (EmDlgItemID) choice;
}


/***********************************************************************
 *
 * FUNCTION:	HostStartIdling
 *
 * DESCRIPTION:	Enable the dialog to receive idle time.
 *
 * PARAMETERS:	context - dialog context data.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmDlg::HostStartIdling (EmDlgContext& context)
{
	// Start idling if we need it.

	if (context.fNeedsIdle)
	{
		::SetTimer ((HWND) context.fDlg, 0, 10, NULL);
	}
}


/***********************************************************************
 *
 * FUNCTION:	HostStopIdling
 *
 * DESCRIPTION:	Disable the dialog from receiving idle time.
 *
 * PARAMETERS:	context - dialog context data.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmDlg::HostStopIdling (EmDlgContext& context)
{
	::KillTimer ((HWND) context.fDlg, 0);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemMin
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemMin (EmDlgRef dlg, EmDlgItemID item, long minValue)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	if (::PrvIsProgressBar (hWnd))
	{
#if 0
		// Can't do this unless _WIN32_IE >= 0x0300.  I'm not sure
		// if that's a condition I want to make.
		PBRANGE	range;
		::SendMessage (hWnd, PBM_GETRANGE, true, (LPARAM) &range);
		range.iLow = minValue;
		::SendMessage (hWnd, PBM_SETRANGE, 0, MAKELPARAM(range.iLow, range.iHigh));
#endif
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemValue
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemValue (EmDlgRef dlg, EmDlgItemID item, long value)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	if (::PrvIsProgressBar (hWnd))
	{
		::SendMessage (hWnd, PBM_SETPOS, value, 0);
	}
	else if (::PrvIsButton (hWnd))
	{
		::SendMessage (hWnd, BM_SETCHECK, value ? BST_CHECKED : BST_UNCHECKED, 0);
	}
	else if (::PrvIsComboBox (hWnd))
	{
		::SendMessage (hWnd, CB_SETCURSEL, value, 0);
	}
	else if (::PrvIsListBox (hWnd))
	{
		::SendMessage (hWnd, LB_SETCURSEL, value, 0);
	}
	else if (::PrvIsEdit (hWnd))
	{
		char	buffer[20];
		sprintf (buffer, "%ld", value);
		::SetWindowText (hWnd, buffer);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemMax
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemMax (EmDlgRef dlg, EmDlgItemID item, long maxValue)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	if (::PrvIsProgressBar (hWnd))
	{
#if 0
		// Can't do this unless _WIN32_IE >= 0x0300.  I'm not sure
		// if that's a condition I want to make.
		PBRANGE	range;
		::SendMessage (hWnd, PBM_GETRANGE, true, (LPARAM) &range);
#endif
		::SendMessage (hWnd, PBM_SETRANGE, 0, MAKELPARAM(0, maxValue));
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemText
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemText (EmDlgRef dlg, EmDlgItemID item, string str)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	// Convert returns to the EOL sequence that Windows likes
	// in a text field.
	string	str2 = ::ReplaceString (str, "\x0A", "\x0D\x0A");

	::EmSetWindowText (hWnd, str2.c_str ());
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetItemValue
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

long EmDlg::GetItemValue (EmDlgRef dlg, EmDlgItemID item)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return 0;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return 0;

	if (::PrvIsProgressBar (hWnd))
	{
#if 0
		// Can't do this unless _WIN32_IE >= 0x0300.  I'm not sure
		// if that's a condition I want to make.
		return (long) ::SendMessage (hWnd, PBM_GETPOS, 0, 0);
#endif
		return 0;
	}
	else if (::PrvIsButton (hWnd))
	{
		return ::SendMessage (hWnd, BM_GETCHECK, 0, 0) == BST_CHECKED ? 1 : 0;
	}
	else if (::PrvIsComboBox (hWnd))
	{
		return (long) ::SendMessage (hWnd, CB_GETCURSEL, 0, 0);
	}
	else if (::PrvIsListBox (hWnd))
	{
		return (long) ::SendMessage (hWnd, LB_GETCURSEL, 0, 0);
	}
	else if (::PrvIsEdit (hWnd))
	{
		char	buffer[20];
		::GetWindowText (hWnd, buffer, 20);
		return (long) atol (buffer);
	}

	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetItemText
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

string EmDlg::GetItemText (EmDlgRef dlg, EmDlgItemID item)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return string();

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return string();

	char	text[256];

	::GetWindowText (hWnd, text, countof(text));

	return string (text);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::EnableItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::EnableItem (EmDlgRef dlg, EmDlgItemID item)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	::EnableWindow (hWnd, true);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::DisableItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::DisableItem (EmDlgRef dlg, EmDlgItemID item)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	::EnableWindow (hWnd, false);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::AppendToMenu
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::AppendToMenu (EmDlgRef dlg, EmDlgItemID item, const StringList& strList)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	if (::PrvIsComboBox (hWnd))
	{
		StringList::const_iterator	iter = strList.begin();
		while (iter != strList.end ())
		{
			::SendMessage (hWnd, CB_ADDSTRING, 0, (LPARAM) iter->c_str ());
			++iter;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::AppendToList
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::AppendToList (EmDlgRef dlg, EmDlgItemID item, const StringList& strList)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	assert (::PrvIsListBox (hWnd));

	StringList::const_iterator	iter = strList.begin();
	while (iter != strList.end ())
	{
		ListBox_AddString (hWnd, iter->c_str ());
		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SelectListItems
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SelectListItems (EmDlgRef dlg, EmDlgItemID item, const EmDlgListIndexList& itemList)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	assert (::PrvIsListBox (hWnd));

	UINT	style = ::GetWindowLong (hWnd, GWL_STYLE);

	int	err;
	
	if ((style & (LBS_MULTIPLESEL | LBS_EXTENDEDSEL)) != 0)
	{
		EmDlgListIndexList::const_iterator	iter = itemList.begin();
		while (iter != itemList.end())
		{
			err = ListBox_SetSel (hWnd, true, *iter);

			++iter;
		}
	}
	else
	{
		err = ListBox_SetCurSel (hWnd, itemList[0]);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetSelectedItems
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::GetSelectedItems (EmDlgRef dlg, EmDlgItemID item, EmDlgListIndexList& itemList)
{
	HWND	hDlg = (HWND) dlg;
	if (!hDlg)
		return;

	HWND	hWnd = ::GetDlgItem (hDlg, ::PrvFromDlgItemID (item));
	if (!hWnd)
		return;

	assert (::PrvIsListBox (hWnd));

	UINT	style = ::GetWindowLong (hWnd, GWL_STYLE);
	
	if ((style & (LBS_MULTIPLESEL | LBS_EXTENDEDSEL)) != 0)
	{
		int	numSelItems = ListBox_GetSelCount (hWnd);

		if (numSelItems > 0)
		{
			int*	selectedItems = new int[numSelItems];
			if (ListBox_GetSelItems (hWnd, numSelItems, selectedItems) != LB_ERR)
			{
				for (int ii = 0; ii < numSelItems; ++ii)
				{
					itemList.push_back (selectedItems[ii]);
				}
			}

			delete [] selectedItems;
		}
	}
	else
	{
		int	sel = ListBox_GetCurSel (hWnd);
		if (sel != LB_ERR)
		{
			itemList.push_back (sel);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	CenterWindow
 *
 * DESCRIPTION:	Centers the given window the desktop window.
 *
 *				(This routine doesn't need to be in this file, but
 *				 there's no better place for it.)
 *
 * PARAMETERS:	hWnd - window to center.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void CenterWindow (HWND hWnd)
{
	RECT	workArea;
	::SystemParametersInfo (SPI_GETWORKAREA, NULL, &workArea, NULL);

	RECT	dRect;
	::GetWindowRect (hWnd, &dRect);

	int xLeft	= (workArea.left + workArea.right) / 2 - (dRect.right - dRect.left) / 2;
	int yTop	= (workArea.top + workArea.bottom) / 2 - (dRect.bottom - dRect.top) / 2;

	::SetWindowPos (hWnd, NULL, xLeft, yTop, -1, -1, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
}


/***********************************************************************
 *
 * FUNCTION:	CenterDlgProc
 *
 * DESCRIPTION:	DlgProc for centering dialogs.
 *
 *				(This routine doesn't need to be in this file, but
 *				 there's no better place for it.)
 *
 * PARAMETERS:	Standard DlgProc parameters.
 *
 * RETURNED:	Standard DlgProc result.
 *
 ***********************************************************************/

UINT CALLBACK CenterDlgProc (	HWND hWnd,
								UINT msg,
								WPARAM wParam,
								LPARAM lParam)
{
	UNUSED_PARAM (wParam);
	UNUSED_PARAM (lParam);

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			::CenterWindow (::GetParent (hWnd));
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	EmSetWindowText
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmSetWindowText(HWND hWndCtrl, const char* lpszNew)
{
	int 	nNewLen = strlen (lpszNew);
	char	szOld[256];

	// fast check to see if text really changes (reduces flash in controls)

	if (nNewLen > countof (szOld) ||
		::GetWindowText (hWndCtrl, szOld, countof(szOld)) != nNewLen ||
		strcmp (szOld, lpszNew) != 0)
	{
		// change it
		::SetWindowText (hWndCtrl, lpszNew);
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvFromDlgID
 *
 * DESCRIPTION:	Convert an EmDlgID to a Windows dialog ID.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

UINT PrvFromDlgID (EmDlgID id)
{
	for (int ii = 0; ii < countof (kDlgIDMap); ++ii)
		if (kDlgIDMap[ii].fMyID == id)
			return kDlgIDMap[ii].fWinID;

	return ID_UNMAPPED;
}


/***********************************************************************
 *
 * FUNCTION:	PrvToDlgID
 *
 * DESCRIPTION:	Convert a Windows dialog ID to an EmDlgID.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

EmDlgID PrvToDlgID (UINT id)
{
	for (int ii = 0; ii < countof (kDlgIDMap); ++ii)
		if (kDlgIDMap[ii].fWinID == id)
			return kDlgIDMap[ii].fMyID;

	return kDlgNone;
}


/***********************************************************************
 *
 * FUNCTION:	PrvFromDlgItemID
 *
 * DESCRIPTION:	Convert an EmDlgItemID to a Windows dialog ID.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

UINT PrvFromDlgItemID (EmDlgItemID id)
{
	for (int ii = 0; ii < countof (kDlgItemIDMap); ++ii)
		if (kDlgItemIDMap[ii].fMyID == id)
			return kDlgItemIDMap[ii].fWinID;

	return ID_UNMAPPED;
}


/***********************************************************************
 *
 * FUNCTION:	PrvToDlgItemID
 *
 * DESCRIPTION:	Convert a Windows dialog ID to an EmDlgItemID.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

EmDlgItemID PrvToDlgItemID (UINT id)
{
	for (int ii = 0; ii < countof (kDlgItemIDMap); ++ii)
		if (kDlgItemIDMap[ii].fWinID == id)
			return kDlgItemIDMap[ii].fMyID;

	return kDlgItemNone;
}


/***********************************************************************
 *
 * FUNCTION:	PrvConvertTypeList
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

FilterList PrvConvertTypeList (const EmFileTypeList& typeList)
{
	FilterList	filter;

	EmFileTypeList::const_iterator	iter = typeList.begin ();

	while (iter != typeList.end ())
	{
		switch (*iter)
		{
			case kFileTypeApplication:
				::PrvAddFilter (filter, "Emulator Executable Files (*.exe)", "*.exe");
				break;

			case kFileTypeROM:
				::PrvAddFilter (filter, "Palm OS ROM Files (*.rom)", "*.rom;* rom;rom.*");
				break;

			case kFileTypeSession:
				::PrvAddFilter (filter, "Emulator Session Files (*.psf)", "*.psf");
				break;

			case kFileTypePalmApp:
				::PrvAddFilter (filter, "Palm OS Applications (*.prc)", "*.prc");
				break;

			case kFileTypePalmDB:
				::PrvAddFilter (filter, "Palm OS Databases (*.pdb)", "*.pdb");
				break;

			case kFileTypePalmQA:
				::PrvAddFilter (filter, "Palm OS Clippings (*.pqa)", "*.pqa");
				break;

			case kFileTypePalmAll:
				::PrvAddFilter (filter, "Palm OS Files (*.prc, *.pdb, *.pqa)", "*.prc;*.pdb;*.pqa");
				::PrvAddFilter (filter, "Palm OS Applications (*.prc)", "*.prc");
				::PrvAddFilter (filter, "Palm OS Databases (*.pdb)", "*.pdb");
				::PrvAddFilter (filter, "Palm OS Clippings (*.pqa)", "*.pqa");
				break;

			case kFileTypePreference:
				::PrvAddFilter (filter, "Preference File (*.ini)", "*.ini");
				break;

			case kFileTypeAll:
				::PrvAddFilter (filter, "All Files (*.*)", "*.*");
				break;
		}

		++iter;
	}

	const char	emptyString[] = "";
	filter.insert (filter.end (), emptyString, emptyString + 1);

	return filter;
}


/***********************************************************************
 *
 * FUNCTION:	PrvAddFilter
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvAddFilter (FilterList& filter, const char* desc, const char* pattern)
{
	filter.insert (filter.end (), desc, desc + strlen (desc) + 1);
	filter.insert (filter.end (), pattern, pattern + strlen (pattern) + 1);
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetStartPosition
 *
 * DESCRIPTION:	Iterator-style function returning the starting
 *				position for the list of files returned from
 *				GetOpenFileName.  Call this function first; then call
 *				GetNextPathName for as long as POSITION is non-zero.
 *
 *				Function unabashedly stolen from MFC.
 *
 * PARAMETERS:	ofn - the OPENFILENAME passed to GetOpenFileName.
 *
 * RETURNED:	An opaque iterator type containing a value indicating
 *				the first item in the list.
 *
 ***********************************************************************/

POSITION PrvGetStartPosition (OPENFILENAME& ofn)
{
	return (POSITION) ofn.lpstrFile;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetNextPathName
 *
 * DESCRIPTION:	Get the pathname indicated by pos, and then increment
 *				pos to refer to the next file in the list.
 *
 *				Function unabashedly stolen from MFC.
 *
 * PARAMETERS:	ofn - the OPENFILENAME passed to GetOpenFileName.
 *
 *				pos - POSITION originally obtained by calling
 *					GetStartPosition.  Subsequent values are returned
 *					by updating pos in-place.
 *
 * RETURNED:	The next pathname indicated by pos.
 *
 ***********************************************************************/

string PrvGetNextPathName (OPENFILENAME& ofn, POSITION& pos)
{
	BOOL	bExplorer	= (ofn.Flags & OFN_EXPLORER) != 0;
	TCHAR	chDelimiter = bExplorer ? '\0' : ' ';
	LPTSTR	lpsz		= (LPTSTR) pos;

	if (lpsz == ofn.lpstrFile) // first time
	{
		if ((ofn.Flags & OFN_ALLOWMULTISELECT) == 0)
		{
			pos = NULL;
			return ofn.lpstrFile;
		}

		// find char pos after first Delimiter
		while (*lpsz != chDelimiter && *lpsz != '\0')
		{
			lpsz = _tcsinc (lpsz);
		}

		lpsz = _tcsinc (lpsz);

		// if single selection then return only selection
		if (*lpsz == 0)
		{
			pos = NULL;
			return ofn.lpstrFile;
		}
	}

	string strPath = ofn.lpstrFile;
	if (!bExplorer)
	{
		LPTSTR	lpszPath = ofn.lpstrFile;

		while (*lpszPath != chDelimiter)
		{
			lpszPath = _tcsinc (lpszPath);
		}

		strPath = strPath.substr (0, lpszPath - ofn.lpstrFile);
	}

	LPTSTR	lpszFileName = lpsz;
	string	strFileName = lpsz;

	// find char pos at next Delimiter
	while (*lpsz != chDelimiter && *lpsz != '\0')
	{
		lpsz = _tcsinc(lpsz);
	}

	if (!bExplorer && *lpsz == '\0')
	{
		pos = NULL;
	}
	else
	{
		if (!bExplorer)
		{
			strFileName = strFileName.substr (0, lpsz - lpszFileName);
		}

		lpsz = _tcsinc (lpsz);

		if (*lpsz == '\0') // if double terminated then done
			pos = NULL;
		else
			pos = (POSITION) lpsz;
	}

	// only add '\\' if it is needed
	if (!strPath.empty())
	{
		// check for last back-slash or forward slash (handles DBCS)
		LPCTSTR lpsz = _tcsrchr (strPath.c_str (), '\\');
		if (lpsz == NULL)
			lpsz = _tcsrchr (strPath.c_str (), '/');

		// if it is also the last character, then we don't need an extra
		if (lpsz != NULL &&
			(lpsz - strPath.c_str ()) == strPath.size () - 1)
		{
			assert (*lpsz == '\\' || *lpsz == '/');
			return strPath + strFileName;
		}
	}

	return strPath + '\\' + strFileName;
}


/***********************************************************************
 *
 * FUNCTION:	PrvOpenMultiFileProc
 *
 * DESCRIPTION:	Hook the OpenFileName dialog so that we can look for
 *				presses on the OK button.  We look for those so that
 *				we can ensure that the buffer receiving the selection
 *				is large enough.  From MSDN:
 *
 *				"With the introduction of the new common dialogs for
 *				Windows 95, a new way of handling the FNERR_BUFFERTOOSMALL
 *				error was developed. It is still necessary to watch for
 *				the Open button to be pressed and reallocate the buffer
 *				if needed, but the way to watch for the OK is much different. 
 *
 *				"When you install a hook on the Open File common dialog
 *				in Windows 95 using the OPENFILENAME.lpfnHook member,
 *				the dialog you are hooking is a child of the main Open
 *				File dialog. Therefore, to intercept the OK button, you
 *				need to subclass the parent dialog. To do this, you can
 *				install the hook procedure and watch for the CDN_INITDONE
 *				notification. The Open File dialog will send this as part
 *				of a WM_NOTIFY message when the initialization for the
 *				dialog is complete."
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

UINT CALLBACK PrvOpenMultiFileProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	UNUSED_PARAM(wParam);

	static HWND hwndParentDialog;
	LPOFNOTIFY lpofn;

	switch (uMsg)
	{
		case WM_INITDIALOG:
			// You need to use a copy of the OPENFILENAME struct used to
			// create this dialog. You can store a pointer to the
			// OPENFILENAME struct in the ofn.lCustData so you can retrieve
			// it here in the lParam. Once you have it, you need to hang on
			// to it. Using window properties provides a good thread safe
			// solution to using a global variable.

			::SetProp(hWnd, "OFN", (HANDLE) lParam);
			::CenterWindow (::GetParent (hWnd));
			return (0);

		case WM_NOTIFY:
			// The OFNOTIFY struct is passed in the lParam of this message.

			lpofn = (LPOFNOTIFY) lParam;

			switch (lpofn->hdr.code)
			{
				case CDN_INITDONE:
					// Subclass the parent dialog to watch for the OK
					// button.

					hwndParentDialog = ::GetParent(hWnd);
					g_lpfnDialogProc = (WNDPROC) ::SetWindowLong(hwndParentDialog,
											DWL_DLGPROC, (LONG) &::PrvOpenMultiFileHook);
					break;

			}
			return (0);

		case WM_DESTROY:
			// Need to clean up the subclassing we did on the dialog.
			::SetWindowLong(hwndParentDialog, DWL_DLGPROC, (LONG) g_lpfnDialogProc);

			// Also need to free the property with the OPENFILENAME struct
			::RemoveProp(hWnd, "OFN");
			return (0);
	}
	return (0);
}


/***********************************************************************
 *
 * FUNCTION:	PrvOpenMultiFileHook
 *
 * DESCRIPTION:	Part 2 of hooking the Open File dialog so that we can
 *				look for clicks on OK.	See OpenMultiFileProc for the
 *				discussion of the first part.  Continuing the MSDN
 *				description:
 *
 *				"Once the parent dialog is subclassed, the program can
 *				watch for the actual Open button. When the program gets
 *				the Open button command, it needs to check to see if the
 *				buffer originally allocated is large enough to handle
 *				all the files selected. The CommDlg_OpenSave_GetFilePath()
 *				API will return the length needed."
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

LRESULT CALLBACK PrvOpenMultiFileHook(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	LPTSTR			lpsz;
	int 			cbLength;
	LPOPENFILENAME	lpofn;

	switch (uMsg)
	{
		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDOK:
					// Need to verify the original buffer size is large
					// enough to handle the files selected. The
					// CommDlg_OpenSave_GetFilePath() API will return the
					// length needed for this buffer.

					cbLength = CommDlg_OpenSave_GetFilePath(hWnd, NULL, 0);

					// MAX_PATH is the size of the buffer originally
					// used in the OPENFILENAME.lpstrFile member.

					if (MAX_PATH < cbLength)
					{
						// The buffer is too small, so allocate a
						// new buffer.
						lpsz = (LPTSTR) malloc (cbLength);
						if (lpsz)
						{
							// The OFN struct is stored in a property of
							// the dialog window.

							lpofn = (LPOPENFILENAME) ::GetProp(hWnd, "OFN");

							free (lpofn->lpstrFile);

							lpofn->lpstrFile = lpsz;
							lpofn->nMaxFile  = cbLength;
						}
					}

					// Now let the dialog handle the message normally.
					break;
			}
			break;
	}

	return (::CallWindowProc(g_lpfnDialogProc, hWnd, uMsg, wParam, lParam));
}


/*
	List of Common Control classes in case we ever need to test for them.

	ANIMATE_ CLASS		Creates animation controls. These controls silently
						display an AVI clip. 
	DATETIMEPICK_CLASS  Creates date and time picker controls. These controls
						provide a simple and intuitive interface to exchange
						date and time information with a user.  
	HOTKEY_CLASS		Creates hotkey controls. These controls make it easy
						for the user to define hotkeys.  
	MONTHCAL_CLASS		Creates month calendar controls. These controls provide
						a simple and intuitive way for a user to select a date
						from a familiar interface.  
	PROGRESS_CLASS		Creates progress bars. These controls indicate the
						progress of a lengthy operation.  
	REBARCLASSNAME		Creates rebar controls. These controls act as a container
						for child windows.  
	STATUSCLASSNAME		Creates status windows. These controls display status
						information in a horizontal window.  
	TOOLBARCLASSNAME	Creates toolbars. These controls contain buttons that
						carry out menu commands.  
	TOOLTIPS_CLASS		Creates tooltip controls. These controls display a small
						pop-up window containing a line of text that describes
						the purpose of a tool in an application.  
	TRACKBAR_CLASS		Creates trackbars. These controls let the user select
						from a range of values by moving a slider.  
	UPDOWN_CLASS		Creates up-down controls. These controls combine a pair
						of arrows with an edit control. Clicking the arrows
						increments or decrements the value in the edit control.  
	WC_COMBOBOXEX		Creates ComboBoxEx controls. These controls provide an
						extension of the combobox control that provides native
						support for item images.  
	WC_HEADER			Creates header controls. These controls display headings
						at the top of columns of information and let the user
						sort the information by clicking the headings.  
	WC_IPADDRESS		Creates IP address controls. These controls are similar
						to an edit control, but allow you to enter a numerical
						address in IP format.  
	WC_LISTVIEW			Creates list view controls. These controls display a
						collection of items, each consisting of an icon and a
						label, and provide several ways to arrange the items.  
	WC_PAGESCROLLER		Creates pager controls. These controls are used to
						contain and scroll another window.  
	WC_TABCONTROL		Creates tab controls. These controls define multiple
						pages for the same area of a window or dialog box. Each
						page consists of a set of information or a group of
						controls that an application displays when the user
						selects the corresponding tab.  
	WC_TREEVIEW			Creates tree view controls. These controls display a
						hierarchical list of items. Each item consists of a
						label and an optional bitmap.  
*/

/*
	List of standard control classes in case we ever need to test for them.

		Listbox
		ComboBox
		ScrollBar
		Button
		Static
		Edit

	and some less familiar controls:

		The class for menus.
		The class for the desktop window.
		The class for dialog boxes.
		The class for the task switch window.
		The class for icon titles.
		ComboLBox: The class for the drop-down list box contained in a combo box.
		MDIClient: The class for multiple-document interface (MDI) client windows.

	also:
		RichEdit		RichEdit 1.0
		RICHEDIT_CLASS	RichEdit 2.0
*/

const char	kButtonClassName[]		= "Button";
const char	kComboBoxClassName[]	= "ComboBox";
const char	kListBoxClassName[]		= "ListBox";
const char	kEditClassName[]		= "Edit";

const UINT	kButtonStyleMask = 0xFF;		// This is what MFC uses in CButton::GetButtonStyle


/***********************************************************************
 *
 * FUNCTION:	PrvIsPushButton
 * FUNCTION:	PrvIsCheckBox
 * FUNCTION:	PrvIsRadioButton
 * FUNCTION:	PrvIsProgressBar
 * FUNCTION:	PrvIsButton
 * FUNCTION:	PrvIsClass
 *
 * DESCRIPTION:	Predicates to test HWNDs to see what kinds of controls
 *				they represent.
 *
 * PARAMETERS:	hWnd - HWND to test
 *
 * RETURNED:	True if the HWND is of the type being tested.
 *
 ***********************************************************************/

Bool PrvIsPushButton (HWND hWnd)
{
	if (!PrvIsButton (hWnd))
		return false;

	UINT	style = ::GetWindowLong (hWnd, GWL_STYLE) & kButtonStyleMask;
	return	style == BS_PUSHBUTTON ||		// Creates a pushbutton that posts a WM_COMMAND message to the owner window when the user selects the button.
			style == BS_DEFPUSHBUTTON;		// Creates a button that has a heavy black border. The user can select this button by pressing the ENTER key. This style enables the user to quickly select the most likely option (the default option).
}

Bool PrvIsCheckBox (HWND hWnd)
{
	if (!PrvIsButton (hWnd))
		return false;

	UINT	style = ::GetWindowLong (hWnd, GWL_STYLE) & kButtonStyleMask;
	return	style == BS_AUTOCHECKBOX ||		// Same as a check box, except that a check mark appears in the check box when the user selects the box; the check mark disappears the next time the user selects the box.
			style == BS_AUTO3STATE ||		// Same as a three-state check box, except that the box changes its state when the user selects it.
			style == BS_CHECKBOX ||			// Creates a small square that has text displayed to its right (unless this style is combined with the BS_LEFTTEXT style).
			style == BS_3STATE;				// Same as a check box, except that the box can be dimmed as well as checked. The dimmed state typically is used to show that a check box has been disabled.
}

Bool PrvIsRadioButton (HWND hWnd)
{
	if (!PrvIsButton (hWnd))
		return false;

	UINT	style = ::GetWindowLong (hWnd, GWL_STYLE) & kButtonStyleMask;
	return	style == BS_AUTORADIOBUTTON ||	// Same as a radio button, except that when the user selects it, the button automatically highlights itself and removes the selection from any other radio buttons with the same style in the same group.
			style == BS_RADIOBUTTON;		// Creates a small circle that has text displayed to its right (unless this style is combined with the BS_LEFTTEXT style). Radio buttons are usually used in groups of related but mutually exclusive choices.
}

Bool PrvIsProgressBar (HWND hWnd)
{
	return PrvIsClass (hWnd, PROGRESS_CLASS);
}

Bool PrvIsButton (HWND hWnd)
{
	return PrvIsClass (hWnd, kButtonClassName);
}

Bool PrvIsComboBox (HWND hWnd)
{
	return PrvIsClass (hWnd, kComboBoxClassName);
}

Bool PrvIsListBox (HWND hWnd)
{
	return PrvIsClass (hWnd, kListBoxClassName);
}

Bool PrvIsEdit (HWND hWnd)
{
	return PrvIsClass (hWnd, kEditClassName);
}

Bool PrvIsClass (HWND hWnd, const char* className)
{
	char	buffer[256];
	if (::GetClassName (hWnd, buffer, countof (buffer)))
	{
		return _stricmp (className, buffer) == 0;	// MFC's _AfxIsComboBoxControl uses an unsigned compare, too.
	}

	return false;
}
