/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "HostPreferences.h"

#include "CPU_REG.h"			// Emulator::ExecuteUntilATrap, Emulator::tracer
#include "EmDlgWin.h"			// CenterWindow, CenterDlgProc
#include "Emulator.h"			// gInstance
#include "EmTransportSerial.h"	// EmTransportSerial::PortNameList, etc.
#include "Hordes.h"				// Hordes::IsOn
#include "Miscellaneous.h"		// SetHotSyncUserName
#include "Platform.h"			// Platform::GetString
#include "PreferenceMgr.h"		// Preferences
#include "Strings.r.h"			// kStr_NoPort
#include "resource.h"			// IDD_PROPERTIES
#include "TracerCommon.h"		// TracerTypeInfo

static void PrvUpdateCheckBox	(HWND w, UINT id, const char* prefName);
static void PrvUpdatePreference (HWND w, UINT id, const char* prefName);
static void PrvUpdateCheckBox	(HWND w, UINT id, const char* prefName, int bitMask);
static void PrvUpdatePreference (HWND w, UINT id, const char* prefName, int bitMask);


// ===========================================================================
//		 HostPreferences
// ===========================================================================

HostPreferences*	gHostPrefs;

// List of user's skin choices.  Used when the Skins dialog is being displayed.

DeviceType		gDeviceChoice;
SkinNameList	gSkinChoices;

unsigned short	gSelectedTracerType;

/***********************************************************************
 *
 * FUNCTION:    HostPreferences::HostPreferences
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

HostPreferences::HostPreferences (void) :
	EmulatorPreferences ()
{
	if (gHostPrefs == NULL)
		gHostPrefs = this;
}


/***********************************************************************
 *
 * FUNCTION:    HostPreferences::~HostPreferences
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

HostPreferences::~HostPreferences (void)
{
	if (gHostPrefs == this)
		gHostPrefs = NULL;
}


/***********************************************************************
 *
 * FUNCTION:    HostPreferences::EditProperties
 *
 * DESCRIPTION: Displays a dialog allowing the user to change the
 *				emulator preferences.  The new settings are saved to
 *				the prefs file.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

static BOOL CALLBACK PrvEditPropertiesDlgProc ( HWND hwnd,
												UINT msg,
												WPARAM wparam,
												LPARAM lparam)
{
	UNUSED_PARAM (lparam);

	string	noPort (Platform::GetString (kStr_NoPort));

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			// Get the preferences we'll be mucking with.

			Preference<string>			prefCommPort (kPrefKeyCommPort);
			Preference<bool>			prefRedirectNetLib (kPrefKeyRedirectNetLib);
			Preference<CloseActionType> prefCloseAction (kPrefKeyCloseAction);
			Preference<string>			prefUserName (kPrefKeyUserName);
			Preference<Configuration>	prefConfiguration (kPrefKeyLastConfiguration);
			Preference<bool>			prefEnableSounds (kPrefKeyEnableSounds);

			::CenterWindow (hwnd);

			HWND	portCombo = ::GetDlgItem (hwnd, IDC_PORT);

			// Add the comm port strings to the port combo box.

			HostPreferences::PopulatePortMenu (portCombo);

			// Add the "no port" menu item.

			ComboBox_AddString (portCombo, noPort.c_str ());

			// Select the menu item for the chosen port.

			string	portName (*prefCommPort);
			if (portName.empty ())
				portName = noPort;

			ComboBox_SelectString (portCombo, -1, portName.c_str ());

			// Set NetLib redirection option

			if (*prefRedirectNetLib)
				::SendDlgItemMessage (hwnd, IDC_NETLIB, BM_SETCHECK, BST_CHECKED, 0);
			else
				::SendDlgItemMessage (hwnd, IDC_NETLIB, BM_SETCHECK, BST_UNCHECKED, 0);

			// Set Enable sounds option

			if (*prefEnableSounds)
				::SendDlgItemMessage (hwnd, IDC_SOUNDS, BM_SETCHECK, BST_CHECKED, 0);
			else
				::SendDlgItemMessage (hwnd, IDC_SOUNDS, BM_SETCHECK, BST_UNCHECKED, 0);

			// Select "Always Save", "Ask Save", or "Never Save".

			if (!::IsBoundFully ())
			{
				if (*prefCloseAction == kSaveAlways)
					::SendDlgItemMessage (hwnd, IDC_SAVEALWAYS, BM_SETCHECK, BST_CHECKED, 0);
				else if (*prefCloseAction == kSaveAsk)
					::SendDlgItemMessage (hwnd, IDC_SAVEASK, BM_SETCHECK, BST_CHECKED, 0);
				else
					::SendDlgItemMessage (hwnd, IDC_SAVENEVER, BM_SETCHECK, BST_CHECKED, 0);
			}

			// Install the user name.

			HWND	userNameItem = ::GetDlgItem (hwnd, IDC_USERNAME);
			::SetWindowText (userNameItem, prefUserName->c_str ());

			break;
		}

		case WM_COMMAND:
		{
			switch (LOWORD (wparam))
			{
				case IDOK:
				{
					// Get the preferences we'll be mucking with.

					Preference<string>			prefCommPort (kPrefKeyCommPort);
					Preference<bool>			prefRedirectNetLib (kPrefKeyRedirectNetLib);
					Preference<CloseActionType> prefCloseAction (kPrefKeyCloseAction);
					Preference<string>			prefUserName (kPrefKeyUserName);
					Preference<Configuration>	prefConfiguration (kPrefKeyLastConfiguration);
					Preference<bool>			prefEnableSounds (kPrefKeyEnableSounds);

					// Get the selected port

					HWND	portCombo = ::GetDlgItem (hwnd, IDC_PORT);
					LONG	index = ComboBox_GetCurSel (portCombo);
					char	tempBuffer[256];
					ComboBox_GetLBText (portCombo, index, tempBuffer);
					prefCommPort = string (tempBuffer);

					if (noPort == *prefCommPort)
						prefCommPort = string ();

					// Get the NetLib redirection setting

					bool	newRedirect = ::SendDlgItemMessage (hwnd, IDC_NETLIB, BM_GETCHECK, 0, 0) == BST_CHECKED;
					prefRedirectNetLib = newRedirect;


					// Get the Enable sounds setting

					bool	newSounds = ::SendDlgItemMessage (hwnd, IDC_SOUNDS, BM_GETCHECK, 0, 0) == BST_CHECKED;
					prefEnableSounds = newSounds;


					// Get the "close action"

					if (!::IsBoundFully ())
					{
						if (::SendDlgItemMessage (hwnd, IDC_SAVEALWAYS, BM_GETCHECK, 0, 0) == BST_CHECKED)
							prefCloseAction = kSaveAlways;
						else if (::SendDlgItemMessage (hwnd, IDC_SAVEASK, BM_GETCHECK, 0, 0) == BST_CHECKED)
							prefCloseAction = kSaveAsk;
						else
							prefCloseAction = kSaveNever;
					}

					// Get and set the specified user name.

					HWND	userNameItem = ::GetDlgItem (hwnd, IDC_USERNAME);
					::GetWindowText (userNameItem, tempBuffer, sizeof (tempBuffer));
					prefUserName = string (tempBuffer);

					::SetHotSyncUserName (tempBuffer);
				}

					// Fall thru

				case IDCANCEL:
					::EndDialog (hwnd, LOWORD (wparam));
					break;
			}
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


void HostPreferences::EditProperties (HWND hwnd)
{
	UINT	resID = ::IsBoundFully () ? IDD_PROPERTIES_FULLY_BOUND : IDD_PROPERTIES;

	int		choice = ::DialogBox (	gInstance, MAKEINTRESOURCE (resID),
									hwnd, &::PrvEditPropertiesDlgProc);
}


/***********************************************************************
 *
 * FUNCTION:    HostPreferences::EditDebugOptions
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

static BOOL CALLBACK PrvEditDebugOptionsDlgProc (	HWND hwnd,
													UINT msg,
													WPARAM wparam,
													LPARAM lparam)
{
	UNUSED_PARAM(lparam)

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			::CenterWindow (hwnd);

			::PrvUpdateCheckBox (hwnd, IDC_FREECHUNK,			kPrefKeyReportFreeChunkAccess);
			::PrvUpdateCheckBox (hwnd, IDC_REGISTERACCESS,		kPrefKeyReportHardwareRegisterAccess);
			::PrvUpdateCheckBox (hwnd, IDC_LOWMEMACCESS,		kPrefKeyReportLowMemoryAccess);
			::PrvUpdateCheckBox (hwnd, IDC_LOWSTACK,			kPrefKeyReportLowStackAccess);
			::PrvUpdateCheckBox (hwnd, IDC_MEMMGRDATA,			kPrefKeyReportMemMgrDataAccess);
			::PrvUpdateCheckBox (hwnd, IDC_MEMMGRSEMAPHORE, 	kPrefKeyReportMemMgrSemaphore);
			::PrvUpdateCheckBox (hwnd, IDC_SCREENACCESS,		kPrefKeyReportScreenAccess);
			::PrvUpdateCheckBox (hwnd, IDC_STACKALMOSTOVERFLOW, kPrefKeyReportStackAlmostOverflow);
			::PrvUpdateCheckBox (hwnd, IDC_STORAGEHEAP, 		kPrefKeyReportStorageHeapAccess);
			::PrvUpdateCheckBox (hwnd, IDC_GLOBALSACCESS,		kPrefKeyReportSystemGlobalAccess);
			::PrvUpdateCheckBox (hwnd, IDC_UNINITCHUNK, 		kPrefKeyReportUninitializedChunkAccess);
			::PrvUpdateCheckBox (hwnd, IDC_UNINITSTACK, 		kPrefKeyReportUninitializedStackAccess);
			::PrvUpdateCheckBox (hwnd, IDC_UNLOCKEDCHUNK,		kPrefKeyReportUnlockedChunkAccess);

			// Disable unsupported options.

			::EnableWindow (GetDlgItem (hwnd, IDC_LOWSTACK), false);
			::EnableWindow (GetDlgItem (hwnd, IDC_STORAGEHEAP), false);
			::EnableWindow (GetDlgItem (hwnd, IDC_UNINITCHUNK), false);
			::EnableWindow (GetDlgItem (hwnd, IDC_UNINITSTACK), false);
			break;
		}

		case WM_COMMAND:
		{
			switch (LOWORD (wparam))
			{
				case IDOK:
					::PrvUpdatePreference (hwnd, IDC_FREECHUNK, 			kPrefKeyReportFreeChunkAccess);
					::PrvUpdatePreference (hwnd, IDC_REGISTERACCESS,		kPrefKeyReportHardwareRegisterAccess);
					::PrvUpdatePreference (hwnd, IDC_LOWMEMACCESS,			kPrefKeyReportLowMemoryAccess);
					::PrvUpdatePreference (hwnd, IDC_LOWSTACK,				kPrefKeyReportLowStackAccess);
					::PrvUpdatePreference (hwnd, IDC_MEMMGRDATA,			kPrefKeyReportMemMgrDataAccess);
					::PrvUpdatePreference (hwnd, IDC_MEMMGRSEMAPHORE,		kPrefKeyReportMemMgrSemaphore);
					::PrvUpdatePreference (hwnd, IDC_SCREENACCESS,			kPrefKeyReportScreenAccess);
					::PrvUpdatePreference (hwnd, IDC_STACKALMOSTOVERFLOW,	kPrefKeyReportStackAlmostOverflow);
					::PrvUpdatePreference (hwnd, IDC_STORAGEHEAP,			kPrefKeyReportStorageHeapAccess);
					::PrvUpdatePreference (hwnd, IDC_GLOBALSACCESS, 		kPrefKeyReportSystemGlobalAccess);
					::PrvUpdatePreference (hwnd, IDC_UNINITCHUNK,			kPrefKeyReportUninitializedChunkAccess);
					::PrvUpdatePreference (hwnd, IDC_UNINITSTACK,			kPrefKeyReportUninitializedStackAccess);
					::PrvUpdatePreference (hwnd, IDC_UNLOCKEDCHUNK, 		kPrefKeyReportUnlockedChunkAccess);

					// Fall thru

				case IDCANCEL:
					::EndDialog (hwnd, LOWORD (wparam));
					break;
			}
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


void HostPreferences::EditDebugOptions (HWND hwnd)
{
	int choice = ::DialogBox (	gInstance, MAKEINTRESOURCE (IDD_DEBUGOPTIONS),
								hwnd, &::PrvEditDebugOptionsDlgProc);
}


/***********************************************************************
 *
 * FUNCTION:    HostPreferences::EditTracingOptions
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void static PrvPopTracerSettings (HWND hwnd)
{
	HWND	tracerParamEdit	= ::GetDlgItem (hwnd, IDC_TRACER_PARAM);
	HWND	tracerAutoConnectBtn= ::GetDlgItem (hwnd, IDC_TRACER_AUTOCONNECT);
	TracerTypeInfo*	curTypeInfo;

	/* Propagate settings from controls to internal table */
	if (gSelectedTracerType != 0)
	{
		curTypeInfo= Emulator::tracer.GetTracerTypeInfo (gSelectedTracerType);
		
		Edit_GetText (tracerParamEdit, curTypeInfo->paramTmpVal, sizeof(curTypeInfo->paramTmpVal));
		curTypeInfo->autoConnectTmpState = (Button_GetCheck(tracerAutoConnectBtn) != 0);
	}
}

void static PrvPushTracerSettings (HWND hwnd)
{
	HWND	tracerTypeComboBox	=	::GetDlgItem (hwnd, IDC_TRACER_TYPE);
	HWND	tracerParamStatic	=	::GetDlgItem (hwnd, IDC_TRACER_PARAM_DESCR);
	HWND	tracerParamEdit	=		::GetDlgItem (hwnd, IDC_TRACER_PARAM);
	HWND	tracerAutoConnectBtn=	::GetDlgItem (hwnd, IDC_TRACER_AUTOCONNECT);
	HWND	tracerStatusStatic=		::GetDlgItem (hwnd, IDC_TRACER_LIBSTATUS);
	TracerTypeInfo* newTracerType;

	/* Propagate settings from internal table to controls */
	if (gSelectedTracerType == 0)
	{
		char dllStatusHint [64];
			
		// Nil output port
		Static_SetText (tracerParamStatic, "");
		ShowWindow (tracerParamEdit, SW_HIDE);
		ShowWindow (tracerStatusStatic, SW_SHOW);
		Emulator::tracer.GetLibraryVersionString (dllStatusHint, sizeof(dllStatusHint));
		Static_SetText (tracerStatusStatic, dllStatusHint);
		EnableWindow (tracerAutoConnectBtn, FALSE);
		Button_SetCheck (tracerAutoConnectBtn, FALSE);
	}
	else
	{
		// Any other tracer type
		newTracerType = Emulator::tracer.GetTracerTypeInfo (gSelectedTracerType);

		Static_SetText (tracerParamStatic, newTracerType->paramDescr);
		Edit_SetText (tracerParamEdit, newTracerType->paramTmpVal);
		ShowWindow (tracerStatusStatic, SW_HIDE);
		ShowWindow (tracerParamEdit, SW_SHOW);
		
		if (newTracerType->autoConnectSupport)
		{
			EnableWindow (tracerAutoConnectBtn, TRUE);
			Button_SetCheck (tracerAutoConnectBtn, newTracerType->autoConnectTmpState);
		}
		else
		{
			Button_SetCheck (tracerAutoConnectBtn, FALSE);
			EnableWindow (tracerAutoConnectBtn, FALSE);
		}
	}
}


static BOOL CALLBACK PrvEditTracingOptionsDlgProc (	HWND hwnd,
													UINT msg,
													WPARAM wparam,
													LPARAM lparam)
{
	UNUSED_PARAM(lparam)
	int i;
	bool parameterValueChanged;

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			HWND	tracerTypeComboBox	= ::GetDlgItem (hwnd, IDC_TRACER_TYPE);
			HWND	tracerParamStatic	= ::GetDlgItem (hwnd, IDC_TRACER_PARAM_DESCR);
			HWND	tracerParamEdit	= ::GetDlgItem (hwnd, IDC_TRACER_PARAM);
			TracerTypeInfo* t;

			// Copy prefs states to internal states
			gSelectedTracerType = Emulator::tracer.GetCurrentTracerTypeIndex();

			for (i= 1; i<=Emulator::tracer.GetTracerTypeCount(); i++)
			{
				t = Emulator::tracer.GetTracerTypeInfo(i);
				strcpy (t->paramTmpVal, t->paramCurVal);
				t->autoConnectTmpState = t->autoConnectCurState;
			}

			// Populate target types
			ComboBox_AddString (tracerTypeComboBox, "None (discards traces)");
			
			for (i=1; i<=Emulator::tracer.GetTracerTypeCount(); i++)
			{
				ComboBox_AddString (tracerTypeComboBox, Emulator::tracer.GetTracerTypeInfo(i)->friendlyName);
			}

			// Select the current one
			ComboBox_SetCurSel (tracerTypeComboBox, gSelectedTracerType);

			PrvPushTracerSettings (hwnd);
			
			// Let the system choose which control will get the focus
			return TRUE; 
		}

		case WM_COMMAND:
		{
			switch (LOWORD (wparam))
			{
                case IDC_TRACER_TYPE:

					switch (HIWORD(wparam))
                    {
						case CBN_SELCHANGE:
						{
							::PrvPopTracerSettings (hwnd);
							gSelectedTracerType = ComboBox_GetCurSel ((HWND)lparam);
							::PrvPushTracerSettings (hwnd);
							return TRUE;
						}
					}
					break;

				case IDOK:
					parameterValueChanged = false;

					if (gSelectedTracerType)
					{
						// Save state for current tracer type only
						::PrvPopTracerSettings (hwnd);
						
						if (strcmp(Emulator::tracer.GetTracerTypeInfo(gSelectedTracerType)->paramCurVal, Emulator::tracer.GetTracerTypeInfo(gSelectedTracerType)->paramTmpVal))
						{
							parameterValueChanged = true;
							strcpy (Emulator::tracer.GetTracerTypeInfo(gSelectedTracerType)->paramCurVal, Emulator::tracer.GetTracerTypeInfo(gSelectedTracerType)->paramTmpVal);
						}

						Emulator::tracer.GetTracerTypeInfo(gSelectedTracerType)->autoConnectCurState = Emulator::tracer.GetTracerTypeInfo(gSelectedTracerType)->autoConnectTmpState;
					}

					// Take the choice into account at the trace engine level
					Emulator::tracer.SetCurrentTracerTypeIndex (gSelectedTracerType, parameterValueChanged);

					// Fall through

				case IDCANCEL:
					::EndDialog (hwnd, LOWORD(wparam));
					return TRUE;
			}
		}

	}

	return FALSE; // The message has not been processed
}


void HostPreferences::EditTracingOptions (HWND hwnd)
{
	int choice = ::DialogBox (	gInstance, MAKEINTRESOURCE (IDD_TRACINGOPTIONS),
								hwnd, &::PrvEditTracingOptionsDlgProc);
}

/***********************************************************************
 *
 * FUNCTION:    HostPreferences::PopulatePortMenu
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void HostPreferences::PopulatePortMenu (HWND portCombo)
{
	EmTransportSerial::PortNameList	portNames;
	EmTransportSerial::GetSerialPortNameList (portNames);
	EmTransportSerial::PortNameList::iterator	iter = portNames.begin ();

	while (iter != portNames.end ())
	{
		ComboBox_AddString (portCombo, iter->c_str ());
		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:    PrvUpdateCheckBox
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void PrvUpdateCheckBox (HWND hwnd, UINT id, const char* prefName)
{
	assert (hwnd);

	Preference<bool>	pref (prefName);

	::SendDlgItemMessage (hwnd, id, BM_SETCHECK, (WPARAM) *pref, 0);
}


/***********************************************************************
 *
 * FUNCTION:    PrvUpdatePreference
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void PrvUpdatePreference (HWND hwnd, UINT id, const char* prefName)
{
	assert (hwnd);

	Preference<bool>	pref (prefName);

	pref = ::SendDlgItemMessage (hwnd, id, BM_GETCHECK, 0, 0) == BST_CHECKED;
}


/***********************************************************************
 *
 * FUNCTION:    PrvUpdateCheckBox
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void PrvUpdateCheckBox (HWND hwnd, UINT id, const char* prefName, int bitMask)
{
	assert (hwnd);

	Preference<uint8>	pref (prefName);

	::SendDlgItemMessage (hwnd, id, BM_SETCHECK, (WPARAM) ((*pref & bitMask) != 0), 0);
}


/***********************************************************************
 *
 * FUNCTION:    PrvUpdatePreference
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void PrvUpdatePreference (HWND hwnd, UINT id, const char* prefName, int bitMask)
{
	assert (hwnd);

	Preference<uint8>	pref (prefName);

	if (::SendDlgItemMessage (hwnd, id, BM_GETCHECK, 0, 0) == BST_CHECKED)
		pref = *pref | bitMask;
	else
		pref = *pref & ~bitMask;
}
